{-# LANGUAGE CPP, MultiParamTypeClasses, FlexibleContexts #-}
module Graphics.SVGFonts.RasterFont
 (raster, withCheckImage, createTexture, texData, bucketSort2d, array2d_to1d,
  bitmask, line, bitSum, endBits, bits,
  F2P, AA(..), Bitmask,
  ) where

import Data.Array.MArray
import Data.Array.ST (STArray,STUArray)
import Data.Array hiding (elems,array)
import Data.Array.Unboxed
import Data.Bits ( (.&.) )
import Data.Word
import Data.Bits (testBit,setBit, shiftL, bit, shiftR)
import Data.STRef
import Control.Monad (when, forM_, forM)
import Control.Monad.ST
import Graphics.Rendering.OpenGL hiding (get,set)
import Foreign (mallocBytes, withArray)
import Debug.Trace
import Graphics.Formats.TGA.TGA
import qualified Data.ByteString as B

var = Data.STRef.newSTRef
get = Data.STRef.readSTRef
set = Data.STRef.writeSTRef
mutate = Data.STRef.modifySTRef

type F2P = (X, Y, AA)
data AA = NB | -- no bitmask
          B Bitmask
type Bitmask = UArray Int Word16
type X = Float
type Y = Float
type F2 = (X,Y)

instance Show AA where
         show (NB) = "NB" -- No Bitmask
         show (B bits) = "B " -- ++ show bits

raster (dx,dy) xs = -- trace (" xs " ++ show (map (\(x,y,_)->(round x, round y)) (rasterr (dx,dy) xs))) $
                    rasterr (dx,dy) xs

rasterr :: (X,Y) -> [F2] -> [F2P]
rasterr _ [a] = []
rasterr (dx,dy) ((p0x,p0y):((p1x,p1y):bs)) | (abs (p1x-p0x)) < dx && (abs (p1y-p0y)) < dy =
                                                div22 ++ rest -- line that is at most one pixel long
                                           | (abs (p1x-p0x)) < dx = -- < 1 && p0x_int == p1x_int =
                                                rasterStraightLine True  ++ rest -- vertical line
                                           | (abs (p1y-p0y)) < dy = -- 1 && p0y_int == p1y_int =
                                                rasterStraightLine False ++ rest -- horizontal line
                                           | otherwise = trace "test" rest -- should not happen
  where  rest = rasterr (dx,dy) ((p1x, p1y):bs)
         (p0x_int, p0y_int) | p0y < p1y = (truncate p0x, truncate p0y)
                            | otherwise = (truncate p1x, truncate p1y)
         (p1x_int, p1y_int) | p0y < p1y = (truncate p1x, truncate p1y)
                            | otherwise = (truncate p0x, truncate p0y)

         rasterStraightLine up | up        = (divide 1 len_y smallerX p0y_int (line 1 len_y pixP0 pixP1)) -- up
                               | otherwise = (divide len_x 1 smallerX p0y_int (line len_x 1 pixP2 pixP3)) -- right
         divide :: Int -> Int -> Int -> Int -> UArray (Int,Int) Word16 -> [F2P]
         divide nx ny x y a = [ (fromIntegral (ex+x), fromIntegral (ey+y), B (mark (ey*nx+ex))) | -- (bitmask ex ey a)) |
		                                                         ey <- [0..(ny-1)], ex <- [0..(nx-1)] ]
         div22 = divide 2 2 p0x_int p0y_int (line 2 2 pixP4 pixP5) -- (array ((0,0),(0,15)) [])
         mark n = array (0,15) [ (i,32) | i <- [0..15] ] -- bitmask ex ey a
         len_x = abs (p1x_int - p0x_int)
         len_y = p1y_int - p0y_int
         smallerX = truncate $ if (p0x < p1x) then p0x else p1x

         pixP0 | p0y < p1y = (truncate $ (frac p0x)*16, truncate $ (frac p0y)*16)
               | otherwise = (truncate $ (frac p1x)*16, truncate $ (frac p1y)*16)
         pixP1 | p0y < p1y = (truncate $ (frac p1x)*16, truncate $ (frac p1y + fromIntegral (p1y_int-p0y_int))*16)
               | otherwise = (truncate $ (frac p0x)*16, truncate $ (frac p0y + fromIntegral (p1y_int-p0y_int))*16)
         pixP2 | p0x < p1x = (truncate $ (frac p0x)*16, truncate $ (frac p0y)*16)
               | otherwise = (truncate $ (frac p1x)*16, truncate $ (frac p1y)*16)
         pixP3 | p0x < p1x = (truncate $ (frac p1x+ fromIntegral (abs (p1x_int-p0x_int)))*16, truncate $ (frac p1y)*16)
               | otherwise = (truncate $ (frac p0x+ fromIntegral (abs (p1x_int-p0x_int)))*16, truncate $ (frac p0y)*16)
         pixP4 | p0y < p1y = (truncate $ (frac p0x)*16, truncate $ (frac p0y)*16)
               | otherwise = (truncate $ (frac p1x)*16, truncate $ (frac p1y)*16)
         pixP5 | p0y < p1y = (truncate $ (frac p1x)*16, truncate $ (frac p1y)*16)
               | otherwise = (truncate $ (frac p0x)*16, truncate $ (frac p0y)*16)

         frac = snd.properFraction

-- fil xs = filter (>0) (elems xs)

texData :: (Int,Int) -> [(Int, Int, Bitmask)] -> TGAData
texData (rx,ry) border_points = tga $ array2d_to1d rx ry $ fillSpans $ (bucketSort2d rx ry border_points)
  where
  tga t = (TGAData (B.empty)
                   (RGB32 (B.concat (map color (elems t))) )
                   0 0
                   rx ry)
  color x = B.cons x $ B.cons x $ B.cons x $ B.singleton x

-- |Fill a bitmask with 1s if the bit/subpixel is inside a polygon, (the bitmask was filled before with
-- subpixels of outline drawing of the polygon). The bitmask is an array of 16 word16 where the
-- x-direction are the bit position in the word16s. Horizontal(x-direction) lines are drawn if the
-- line is inside the polygon, called a span (see Jordan curve theorem).
-- A subpixel marks the beginnig or the end of a span depending on the state of the line.
-- At the beginning all lines are white, if one intersects a subpixel it becomes black until it
-- hits a subpixel again. The states at the end of the lines are stored for the next pixel in fillSpans.
-- If these state bits are all one or all zero and the next bitmask is completely zero then fewer
-- calculations need to be done.

  fillSpans :: (UArray (Int,Int) Bool, UArray (Int,Int) Word16) -> UArray (Int,Int) Word16
  fillSpans (ar_bool, ar_word) = runST ( -- trace ("span " ++ show (fil ar_word)) $ runST (
    do ar <- newArray ((0,0),(rx-1,ry-1)) 0 :: ST s (STUArray s (Int,Int) Word16)
       arb <- thaw2 ar_bool
       arw  <- thaw2 ar_word
       forM_ [0..(ry-1)] ( \y -> do
         startBits <- var (0::Word16)
         forM_ [0..(rx-1)] ( \x -> do
           b <- readArray arb (x,y)
           sb <- get startBits
           case b of True -> do s <- var $ False
                                forM_ [0..15] ( \suby -> do
                                  set s (testBit sb suby)
                                  forM_ [0..15] ( \subx -> do
                                    line <- readArray arw (x,y*16+suby)
                                    let b = (testBit line subx)
                                    t <- get s
                                    when (t==False && b == True) (writeArray arw (x,y*16+suby) (setBit line subx))
                                    when (t==False && b == True) (set s True)                                       -- start filling
                                    when (t==True  && b == False) (writeArray arw (x,y*16+suby) (setBit line subx)) -- continue
                                    when (t==True  && b == True) (set s False) ) )                                  -- stop
                                bitm <- unsafeFreeze arw
                                let bsum = bitSum (bitmask x y bitm)       -- trace (show ((bitmask x y bitm))) $
                                set startBits (endBits (bitmask x y bitm)) -- ((endBits newMask) && not (endBits bm) )
                                writeArray ar (x,y) bsum
                     False -> do writeArray ar (x,y) (shiftL (bits sb) 4) )) -- the same as: (bits startbits)*16
       a' <- unsafeFreeze ar
       return (a' :: UArray (Int,Int) Word16) )


withCheckImage :: [Word8] -> TextureSize2D -> ([Word8] -> [(Color4 GLubyte)]) -> (PixelData (Color4 GLubyte) -> IO ()) -> IO ()
withCheckImage xs (TextureSize2D w h) f act =
  withArray (f xs) -- (map f (map fromIntegral xs))
  $ act. PixelData RGBA UnsignedByte

createTexture :: (Int,Int) -> TGAData -> String -> IO (TextureObject, String)
createTexture (rx,ry) (TGAData _ (RGB32 picture) _ _ w h) fileName = tex (B.unpack picture)
  where
  tex :: [Word8] -> IO (TextureObject,String)
  tex xs = do
    [texName] <- genObjectNames 1
    textureBinding Texture2D $= Just texName
    textureFilter Texture2D $= ((Nearest, Nothing), Nearest)
    let imageSize = TextureSize2D (fromIntegral rx) (fromIntegral ry)
    withCheckImage xs imageSize f $ -- (\c -> Color4 c c c 255) $
      texImage2D Nothing NoProxy 0  RGBA' imageSize 0
    return (texName, fileName)

  f (a:(b:(c:(d:ds)))) = (Color4 (fromIntegral a) (fromIntegral b) (fromIntegral c) 255) : (f ds)
  f _ = []

-- type Bitmask = UArray Int Word16

-- |generate a 2-dimensional array in which pixels are inserted
bucketSort2d :: Int -> Int -> [(Int,Int,Bitmask)] -> (UArray (Int,Int) Bool, UArray (Int,Int) Word16)
bucketSort2d rx ry pixels = -- trace ("bucket " ++ show pixels) -- (map (\(x,y,z) -> (x,y)) pixels) )
 (runST ( do
  ar0 <- newArray ((0,0),(rx-1,ry-1)) False :: ST s (STUArray s (Int,Int) Bool) -- == True if pixel has a grey value (the outline goes through it)
  ar1 <- newArray ((0,0),(rx-1,ry*16-1)) 0 :: ST s (STUArray s (Int,Int) Word16)-- an unboxed array of unboxed arrays is simulated with one big array
  forM pixels (\(ix,iy,mask) -> do
    when (ix >= 0 && iy >= 0 && (ix+iy*rx) < rx*ry) ( do
      bbool <- readArray ar0 (ix, iy)
      bm1 <- thawSTU mask
      case bbool of False -> do writeArray ar0 (ix,iy) True
                                forM_ [0..15] ( \y -> do
                                  e <- readArray bm1 y
                                  writeArray ar1 (ix,iy*16+y) e )
                    True  -> do writeArray ar0 (0,0) True
--                                forM_ [0..15] ( \y -> do
--                                  e0 <- readArray ar1 (ix,iy*16+y)
--                                  e1 <- readArray bm1 y
--                                  writeArray ar1 (ix,iy*16+y) (e0 `xor` e1) )
                   ) )
  a0 <- unsafeFreeze ar0
  a1 <- unsafeFreeze ar1
  return (a0 :: UArray (Int,Int) Bool,  a1 :: UArray (Int,Int) Word16) ))

ro = (\(x,y,z)->(round x, round y, z))

thawSTU :: (IArray UArray e, MArray (STUArray s) e (ST s)) => UArray Int e -> ST s (STUArray s Int e)
thawSTU = thaw

thaw2 :: (IArray UArray e, MArray (STUArray s) e (ST s)) => UArray (Int,Int) e -> ST s (STUArray s (Int,Int) e)
thaw2 = thaw

array2d_to1d :: Int -> Int -> UArray (Int,Int) Word16 ->  UArray Int Word8
array2d_to1d rx ry mask = runST $ do -- trace ("mask " ++ show mask) runST $ do
  ar <- thaw2 mask
  m <- newArray (0,rx*ry-1) (0::Word8) :: ST s (STUArray s Int Word8)
  forM_ [0..(ry-1)] ( \y -> do
    forM_ [0..(rx-1)] ( \x -> do
      e <- readArray ar (x,y)
      writeArray m (x+y*rx) (fromIntegral e)
                ) )
  m' <- unsafeFreeze m
  return (m' :: UArray Int Word8)

-- |slice out a single bitmask from an array of bitmasks
-- (this should be an unboxed array of unboxed arrays,
--  but since this is not so easy in Haskell, one 2d-array with a supersampled y-coordinate (*16) is used)
-- the x-supersampling are the bits in the Word16
bitmask :: Int -> Int -> UArray (Int,Int) Word16 -> Bitmask
bitmask ix iy mask = runST $ do -- trace ("x,y "++ show ix ++ "," ++ show iy ++ " ar ") (runST $ do
  ar <- thaw2 mask
  m <- newArray (0,15) (0::Word16) :: ST s (STUArray s Int Word16)
  forM_ [iy*16..iy*16+15] $ \y -> do
    e <- readArray ar (ix,y)
    writeArray m (y-iy*16) e
  m' <- unsafeFreeze m
  return (m' :: Bitmask)

-- | bresenham line algorithm
-- the data structure for the line is a 2d array of bits, in x-direction a sequence of word16s
-- example: 
line :: Int -> Int -> (Int,Int) -> (Int,Int) -> UArray (Int,Int) Word16
line nx ny (xa, ya) (xb, yb) = runST $ do  -- nx,ny are the size as multiples of 16,  draw line from (xa, ya) to (xb, yb)
    a <- newArray ((0,0),(nx-1,ny*16-1)) (0 :: Word16) :: ST s (STUArray s (Int,Int) Word16)
    yV <- var y1
    errorV <- var $ deltax `div` 2
    forM_ [x1 .. x2] (\x -> do
        y <- get yV
        setPix $ if steep then (a, y, x) else (a, x, y)
        mutate errorV $ subtract deltay
        error <- get errorV
        when (error < 0) (do
            mutate yV (+ ystep)
            mutate errorV (+ deltax)))
    a' <- unsafeFreeze a
    return (a' :: UArray (Int,Int) Word16)
  where steep = abs (yb - ya) > abs (xb - xa)
        (xa', ya', xb', yb') = if steep
          then (ya, xa, yb, xb)
          else (xa, ya, xb, yb)
        (x1, y1, x2, y2) = if xa' > xb'
          then (xb', yb', xa', ya')
          else (xa', ya', xb', yb')
        deltax = x2 - x1
        deltay = abs $ y2 - y1
        ystep = if y1 < y2 then 1 else -1
        setPix (a, x, y) = when (x<nx*16 && y<ny*16 && x>=0 && y>=0)
                          (writeArray a ((x `div` 16), y) (setBit (0::Word16) (x `mod` 16))) -- (x `mod` 16) can be replaced by x?

-- data Bitmask = UArray Int Word16
-- | the sum of all bits in a Bitmask, 16 rows with 16 bits each => between 0 and 255
bitSum :: Bitmask -> Word16
bitSum mask = runST ( do
  s <- var $ 0
  ar <- thawSTU mask
  forM_ [0..15] $ \y -> do
    e <- readArray ar y
    mutate s (+ bits e)
  get s)

-- | return the the rightmost column of a bitmask
endBits :: Bitmask -> Word16
endBits mask = runST ( do
  s <- var $ 0
  ar <- thawSTU mask
  forM_ [0..15] $ \y -> do
    e <- readArray ar y
    mutate s (+ shiftL (e .&. (bit 15)) y)
  get s)

-- | count the number of 1-bits with divide and conquer
-- it can be done a little bit faster, but for the beginning it should just be correct
-- see "Hacker's Delight by Henry S. Warren, Addison Wesley" for bit counting
bits :: Word16 -> Word16
bits w16 = runST ( do
  s <- var $ w16
  -- fst line: starting to count the 1's in 2-tuples: 0x55 = 01010101
  -- 2nd line: 4 tuples:  0x33 = 0011001100
  sv <- get s
  set s (sv .&. 0x5555 + (shiftR sv 1) .&. 0x5555)
  sv <- get s
  set s (sv .&. 0x3333 + (shiftR sv 2) .&. 0x3333)
  sv <- get s
  set s (sv .&. 0x0F0F + (shiftR sv 4) .&. 0x0F0F)
  sv <- get s
  set s (sv .&. 0x00FF + (shiftR sv 8))
  get s )
